import json
import pandas as pd
import numpy as np
from pathlib import Path
import os
import logging

from cachetools import cached, TTLCache

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

BASE_PATH = Path(__file__).resolve().parent

@cached(cache=TTLCache(maxsize=1024, ttl=1800))
def get_ewr_calc_config(file_path:str = None) -> dict:
    '''Loads the ewr calculation configuration file from repository or local file
    system
    
    Args:
        file_path (str): Location of the ewr calculation configuration file

    Returns:
        dict: Returns a dictionary of the ewr calculation configuration file
    '''
    
    if file_path:
        with open(file_path, 'r') as fp:
            ewr_calc_config = json.load(fp)
    
    if not file_path:
        repo_path = os.path.join(BASE_PATH, "parameter_metadata/ewr_calc_config.json")
        with open(repo_path, 'r') as fp:
            ewr_calc_config = json.load(fp)

    return ewr_calc_config

def modify_EWR_table(EWR_table:pd.DataFrame) -> pd.DataFrame:
  
    ''' 
        Splits StartMonth and EndMonth floats into 4 columns containing integers of StartMonth and StartDay, EndMonth and EndDay
        Coerces columns that are members of the all_int_params list to Int64 and columns that are members of all_float_params to Float64 dtype.
    '''

    all_int_params = ['FlowThresholdMin', 'FlowThresholdMax', 'VolumeThreshold', 
                    'Duration', 'WithinEventGapTolerance', 'EventsPerYear', 'MinSpell', 
                    'AccumulationPeriod', 'MaxSpell', 'TriggerDay', 'TriggerMonth', 'AnnualBarrageFlow', 
                    'ThreeYearsBarrageFlow', 'HighReleaseWindowStart', 'HighReleaseWindowEnd', 'LowReleaseWindowStart', 
                    'LowReleaseWindowEnd', 'PeakLevelWindowStart', 'PeakLevelWindowEnd', 'LowLevelWindowStart', 
                    'LowLevelWindowEnd', 'NonFlowSpell', 'EggsDaysSpell', 'LarvaeDaysSpell', 
                    'StartDay', 'EndDay', 'StartMonth', 'EndMonth']

    all_float_params = ['RateOfRiseMax1', 'RateOfRiseMax2', 'RateOfFallMin', 
                        'RateOfRiseThreshold1', 'RateOfRiseThreshold2', 'RateOfRiseRiverLevel', 
                        'RateOfFallRiverLevel', 'CtfThreshold', 'MaxLevelChange', 'LevelThresholdMin', 
                        'LevelThresholdMax', 'DrawDownRateWeek', 'MaxInter-event']

    months = ['StartMonth', 'EndMonth']

    # Modify startmonth/endmonth
    for col_name in months:
      rows = EWR_table[col_name].copy().items()
      day_col_name = col_name[:-5]+"Day"
      for r_idx, val in rows:
        if "." in val:
          month, day = val.split('.')
        else:
          month = val
          day = None
        EWR_table.loc[r_idx, col_name] = month
        EWR_table.loc[r_idx, day_col_name] = day 
    #NOTE: I actually think the drawdown rate modifications were doing nothing and the handling of percentage / float values is done in all functions that use drawdown_rate.
    float_params = [col for col in EWR_table.columns if col in all_float_params]
    int_params = [col for col in EWR_table.columns if col in all_int_params]
    # Modify floats
    for col_name in float_params:
        EWR_table[col_name] = pd.to_numeric(EWR_table[col_name], errors='coerce').astype('Float64')

    # Modify integers
    for col_name in int_params:
        EWR_table[col_name] = pd.to_numeric(EWR_table[col_name], errors='coerce').astype('Int64')

    return EWR_table

@cached(cache=TTLCache(maxsize=1024, ttl=1800))

def get_EWR_table(file_path:str = None, columns_to_keep = None) -> dict:
    
    ''' 
        Loads internal parameter sheet or parameter sheet from a specified path with user defined or default desired columns.
        Handles conversion of '' values to a maximum value for parameters that are required to have a maximum value,
        that are otherwise not defined in their respective Long Term Water Plans
        
    
    Args:
        file_path (str): Location of the ewr dataset
        columns_to_keep (None |List): list of columns to keep, if None, default list of columns selected. 
        --------------
        NOTE: For EWR tool runs, it is desierable to use the default 'columns_to_keep = None' to ensure correct calcualtion and prevent errors,
        however this option is provided for other uses of the parameter sheet such as cleaning and testing with internal test functions, other analyses, and general viewing. 
        --------------
    Returns:
        tuple(pd.DataFrame, pd.DataFrame): EWRs that meet the minimum requirements; EWRs that dont meet the minimum requirements
    '''
    if columns_to_keep is None:
        cols_keep = ['PlanningUnitID', 'PlanningUnitName', 'Gauge', 'Code', 'StartMonth', 'TargetFrequency', 'State', 'SWSDLName',
                          'EndMonth', 'EventsPerYear', 'Duration', 'MinSpell', 
                          'FlowThresholdMin', 'FlowThresholdMax', 'MaxInter-event', 'WithinEventGapTolerance', 'WeirpoolGauge', 'FlowLevelVolume', 
                          'LevelThresholdMin', 'LevelThresholdMax', 'VolumeThreshold', 'DrawdownRate', 'MaxLevelChange', 'AccumulationPeriod',
                          'Multigauge', 'MaxSpell', 'TriggerDay', 'TriggerMonth', 'DrawDownRateWeek','AnnualBarrageFlow',
                          'ThreeYearsBarrageFlow', 'HighReleaseWindowStart', 'HighReleaseWindowEnd', 'LowReleaseWindowStart', 'LowReleaseWindowEnd',
                          'PeakLevelWindowStart', 'PeakLevelWindowEnd', 'LowLevelWindowStart', 'LowLevelWindowEnd', 'NonFlowSpell','EggsDaysSpell',
                          'LarvaeDaysSpell', 'RateOfRiseMax1','RateOfRiseMax2','RateOfFallMin','RateOfRiseThreshold1',
                          'RateOfRiseThreshold2','RateOfRiseRiverLevel','RateOfFallRiverLevel', 'CtfThreshold', 'GaugeType']
        
    else:
        cols_keep = columns_to_keep
    
    if file_path:
      my_url = file_path
    else:
      my_url = os.path.join(BASE_PATH, "parameter_metadata/parameter_sheet.csv")
    df = pd.read_csv(my_url, usecols = cols_keep, dtype='str', encoding='cp1252')
    clean_df = df.copy(deep = True)
    clean_df = df.fillna('')
    # Here are all the prior assumptions of what to fill in to the parameter sheet if the value is missing.
    # The aim is to remove all of these and have the parameter sheet be correct, the tool should not run
    # the calculation of an ewr with missing (or extra?) values.
    essential_cols = ['StartMonth','EndMonth', 'FlowThresholdMax','LevelThresholdMax',
                      'FlowThresholdMin', 'MaxInter-event', 'WithinEventGapTolerance',
                      'CtfThreshold', 'NonFlowSpell', 'DrawdownRate', 'DrawDownRateWeek', 'MaxSpell', 'MaxLevelChange']
    
    missing_cols = [col for col in essential_cols if col not in df.columns]
    if missing_cols:
        raise KeyError(f"Parameter sheet must contain these columns: {missing_cols}")
    

    clean_df.loc[:, 'FlowThresholdMax'] = (clean_df['FlowThresholdMax'].replace('', '1000000'))
    clean_df.loc[:, 'LevelThresholdMax'] = (clean_df['LevelThresholdMax'].replace('', '1000000'))
    clean_df.loc[:, 'FlowThresholdMin'] = (clean_df['FlowThresholdMin'].replace('', '0'))
    clean_df.loc[:, 'LevelThresholdMin'] = (clean_df['LevelThresholdMin'].replace('', '0'))
    clean_df.loc[:, 'MaxInter-event'] = (clean_df['MaxInter-event'].replace('', '0'))
    clean_df.loc[:, 'WithinEventGapTolerance'] = (clean_df['WithinEventGapTolerance'].replace('', '0'))

    clean_df.loc[:, 'CtfThreshold'] = (clean_df['CtfThreshold'].replace('', '5'))
    clean_df.loc[:, 'NonFlowSpell'] = (clean_df['NonFlowSpell'].replace('', '0'))
    clean_df.loc[:, 'DrawDownRateWeek'] = (clean_df['DrawDownRateWeek'].replace('30', '0.03'))
    clean_df.loc[:, 'DrawDownRateWeek'] = (clean_df['DrawDownRateWeek'].replace('30%', '0.03'))#just for tes t, change the PS in that test to reflect this
    clean_df.loc[:, 'DrawdownRate'] = (clean_df['DrawdownRate'].replace('', '1000000'))
    clean_df.loc[:, 'MaxSpell'] = (clean_df['MaxSpell'].replace('', '1000000'))
    clean_df.loc[:, 'MaxLevelChange'] = (clean_df['MaxLevelChange'].replace('', '1000000'))

    mod_df = modify_EWR_table(clean_df)
    return mod_df

def get_components_map() -> dict:
    components_map = {
        'PlanningUnitID': 'PlanningUnit',
        'Gauge': 'Gauge',
        'Code': 'Code',
        'FlowThresholdMin': 'min_flow',
        'FlowThresholdMax': 'max_flow',
        'VolumeThreshold': 'min_volume',
        'Duration': 'duration',
        'WithinEventGapTolerance': 'gap_tolerance',
        'EventsPerYear': 'events_per_year',
        'MinSpell': 'min_event',
        'AccumulationPeriod': 'accumulation_period',
        'MaxSpell': 'max_duration',
        'TriggerDay': 'trigger_day',
        'TriggerMonth': 'trigger_month',
        'AnnualBarrageFlow': 'annual_barrage_flow',
        'ThreeYearsBarrageFlow': 'three_years_barrage_flow',
        'HighReleaseWindowStart': 'high_release_window_start',
        'HighReleaseWindowEnd': 'high_release_window_end',
        'LowReleaseWindowStart': 'low_release_window_start',
        'LowReleaseWindowEnd': 'low_release_window_end',
        'PeakLevelWindowStart': 'peak_level_window_start',
        'PeakLevelWindowEnd': 'peak_level_window_end',
        'LowLevelWindowStart': 'low_level_window_start',
        'LowLevelWindowEnd': 'low_level_window_end',
        'NonFlowSpell': 'non_flow_spell',
        'EggsDaysSpell': 'eggs_days_spell',
        'LarvaeDaysSpell': 'larvae_days_spell',
        'RateOfRiseMax1': 'rate_of_rise_max1',
        'RateOfRiseMax2': 'rate_of_rise_max2',
        'RateOfFallMin': 'rate_of_fall_min',
        'RateOfRiseThreshold1': 'rate_of_rise_threshold1',
        'RateOfRiseThreshold2': 'rate_of_rise_threshold2',
        'RateOfRiseRiverLevel': 'rate_of_rise_river_level',
        'RateOfFallRiverLevel': 'rate_of_fall_river_level',
        'CtfThreshold': 'ctf_threshold',
        'MaxLevelChange': 'max_level_change',
        'LevelThresholdMin': 'min_level',
        'LevelThresholdMax': 'max_level',
        'StartMonth': 'start_month',
        'EndMonth': 'end_month',
        'StartDay': 'start_day',
        'EndDay': 'end_day',
        'DrawDownRateWeek': 'drawdown_rate_week',
        'MaxInter-event': 'max_inter-event',
        'Multigauge': 'second_gauge',
        'DrawdownRate': 'drawdown_rate',
        'FlowLevelVolume': 'flow_level_volume',
        'WeirpoolGauge': 'weirpool_gauge',
        
        # these are required at the end after scenarios are processed
        # TODO: Clarify/check if we require more here for interevents, etc. other 4 outputs of the tool
        'TargetFrequency': 'TargetFrequency',
        'PlanningUnitName': 'PlanningUnitName',
        'State': 'State', 
        'SWSDLName': 'SWSDLName',
        'GaugeType': 'GaugeType',
        }
    return components_map

def get_MDBA_codes(model_type: str) -> pd.DataFrame:
    '''
    Load MDBA model metadata file containing model nodes
    and gauges they correspond to

    Returns:
        pd.DataFrame: dataframe for linking MDBA model nodes to gauges

    '''
    if model_type == 'Bigmod - MDBA':
        metadata = pd.read_csv( BASE_PATH / 'model_metadata/SiteID_MDBA.csv', engine = 'python', dtype=str)#, encoding='windows-1252')
    if model_type == 'FIRM - MDBA':
        metadata = pd.read_csv( BASE_PATH / 'model_metadata/EWR_Sitelist_FIRM_20250718.csv', engine = 'python', dtype=str)

    return metadata
  
def get_NSW_codes() -> pd.DataFrame:
    '''
    Load NSW model metadata file containing model nodes
    and gauges they correspond to

    Returns:
        pd.DataFrame: dataframe for linking NSW model nodes to gauges

    '''
    metadata = pd.read_csv( BASE_PATH / 'model_metadata/SiteID_NSW.csv', engine = 'python', dtype=str)
    
    return metadata

def get_iqqm_codes() -> dict:
    '''
    Load metadata file for Macquarie containing model nodes
    and gauges they correspond to

    Returns:
        dict: dict for linking model nodes to gauges
    '''

    metadf = pd.read_csv( BASE_PATH / 'model_metadata/iqqm_stations.csv', dtype=str)
    metadata = metadf.set_index(metadf.columns[0]).to_dict()[metadf.columns[1]]
    return metadata

def get_level_gauges() -> tuple:
    '''Returning level gauges with EWRs

    Returns:
        tuple[list, dict]: list of lake level gauges, dictionary of weirpool gauges
    
    '''
    
    menindeeGauges = ['425020', '425022', '425023']

    lachlanGauges = ['412107']

    levelGauges = menindeeGauges + lachlanGauges 
    
    weirpoolGauges = {'414203': '414209', 
                      '425010': 'A4260501',
                      'A4260507': 'A4260508',
                      'A4260505': 'A4260506'}
    
    return levelGauges, weirpoolGauges


def get_multi_gauges(dataType: str) -> dict:
    '''
    Multi gauges are for EWRs that require the flow of two gauges to be added together

    Args:
        dataType (str): Pass 'all' to get multi gauges and their planning units, pass 'gauges' to get only the gauges.
    Returns:
        dict: if 'all' nested dict returned with a level for planning units
    '''
    
    all = {'PU_0000130': {'421090': '421088', '421088': '421090'},
              'PU_0000131': {'421090': '421088', '421088': '421090'},
              'PU_0000132': {'421090': '421088', '421088': '421090'},
              'PU_0000133': {'421090': '421088', '421088': '421090'},
              'PU_0000251': {'423001': '423002'},
              'PU_0000280': {'1AS': '1ES'}
             }
    returnData = {}
    if dataType == 'all':
        returnData = all
    if dataType == 'gauges':
        for i in all:
            returnData = {**returnData, **all[i]}
    
    return returnData

def get_EWR_components(category):
    '''
    Ingests ewr category, returns the components required to analyse this type of ewr. 
    Each code represents a unique component in the ewr dataset.

    Args:
        category (str): options =   'flow', 'low flow', 'cease to flow', 'cumulative', 'level', 'weirpool-raising', 'weirpool-falling', 'nest-level', 'nest-percent',
                                    'multi-gauge-flow', 'multi-gauge-low flow', 'multi-gauge-cease to flow', 'multi-gauge-cease to flow', 'multi-gauge-cumulative', 
                                    'simul-gauge-flow', 'simul-gauge-low flow', 'simul-gauge-cease to flow', 'complex'

    Returns:
        list: Components needing to be pulled from the ewr dataset
    '''

    if category == 'flow':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration', 'MinSpell', 'WithinEventGapTolerance', 'EventsPerYear', 'MaxInter-event', 'FlowLevelVolume']
    elif category == 'low flow':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration',  'MinSpell', 'EventsPerYear', 'MaxInter-event', 'FlowLevelVolume']
    elif category == 'cease to flow':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration', 'MinSpell', 'EventsPerYear', 'MaxInter-event', 'FlowLevelVolume']
    elif category == 'cumulative':
        pull =  ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'VolumeThreshold', 'EventsPerYear', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration', 'MinSpell','MaxInter-event','AccumulationPeriod','WithinEventGapTolerance', 'FlowLevelVolume']
    elif category == 'cumulative_bbr':
        pull =  ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'VolumeThreshold', 'EventsPerYear', 'FlowThresholdMin', 'FlowThresholdMax','Duration', 'MinSpell', 'MaxInter-event','AccumulationPeriod','WithinEventGapTolerance', 'FlowLevelVolume','LevelThresholdMax','WeirpoolGauge']
    elif category == 'water_stability':
        pull =  ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'Duration', 'MinSpell', 'EventsPerYear', 'FlowThresholdMin', 'FlowThresholdMax', 'MaxInter-event','AccumulationPeriod','WithinEventGapTolerance', 'FlowLevelVolume','LevelThresholdMax','WeirpoolGauge', 'EggsDaysSpell', 'LarvaeDaysSpell', 'MaxLevelChange', 'DrawdownRate']
    elif category == 'water_stability_level':
        pull =  ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'Duration', 'MinSpell', 'EventsPerYear', 'FlowThresholdMin', 'MaxInter-event','AccumulationPeriod','WithinEventGapTolerance', 'FlowLevelVolume','LevelThresholdMax', 'LevelThresholdMin', 'WeirpoolGauge', 'EggsDaysSpell', 'LarvaeDaysSpell', 'MaxLevelChange', 'DrawdownRate']
    elif category == 'level':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'LevelThresholdMin', 'LevelThresholdMax', 'Duration', 'MinSpell', 'EventsPerYear', 'DrawdownRate', 'MaxInter-event', 'FlowLevelVolume', 'MaxSpell','WithinEventGapTolerance']
    elif category == 'weirpool-raising':
        pull=['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'LevelThresholdMin', 'Duration', 'MinSpell',  'DrawdownRate', 'EventsPerYear','WeirpoolGauge', 'MaxInter-event', 'FlowLevelVolume', 'WithinEventGapTolerance']
    elif category == 'weirpool-falling':
        pull=['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'LevelThresholdMax', 'Duration', 'MinSpell',  'DrawdownRate', 'EventsPerYear','WeirpoolGauge', 'MaxInter-event', 'FlowLevelVolume', 'WithinEventGapTolerance']
    elif category == 'nest-level':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration', 'MinSpell', 'EventsPerYear', 'WeirpoolGauge', 'MaxInter-event', 'FlowLevelVolume','DrawDownRateWeek','WithinEventGapTolerance']
    elif category == 'nest-percent':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration', 'MinSpell',  'DrawdownRate', 'EventsPerYear', 'MaxInter-event', 'FlowLevelVolume','TriggerDay','TriggerMonth','WithinEventGapTolerance']
    elif category == 'multi-gauge-flow':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration', 'MinSpell',  'WithinEventGapTolerance', 'EventsPerYear', 'Multigauge', 'MaxInter-event', 'FlowLevelVolume']
    elif category == 'multi-gauge-low flow':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration', 'MinSpell', 'EventsPerYear', 'Multigauge', 'MaxInter-event', 'FlowLevelVolume']
    elif category == 'multi-gauge-cease to flow':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration', 'MinSpell', 'EventsPerYear', 'Multigauge', 'MaxInter-event', 'FlowLevelVolume']
    elif category == 'multi-gauge-cumulative':
        pull =  ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'VolumeThreshold', 'EventsPerYear', 'FlowThresholdMin', 'FlowThresholdMax','Duration', 'MinSpell','Multigauge', 'MaxInter-event','AccumulationPeriod','WithinEventGapTolerance', 'FlowLevelVolume']
    elif category == 'flood-plains':
        pull=['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'LevelThresholdMax', 'Duration', 'MinSpell',  'DrawdownRate', 'MaxLevelChange','EventsPerYear','WeirpoolGauge', 'MaxInter-event', 'FlowLevelVolume', 'WithinEventGapTolerance']
    elif category == 'barrage-flow':
        pull=['StartMonth', 'EndMonth', 'StartDay', 'EndDay','Duration', 'MinSpell','EventsPerYear','MaxInter-event','FlowLevelVolume','AnnualBarrageFlow','ThreeYearsBarrageFlow','HighReleaseWindowStart', 'HighReleaseWindowEnd', 'LowReleaseWindowStart', 'LowReleaseWindowEnd']
    elif category == 'barrage-level':
        pull=['StartMonth', 'EndMonth', 'StartDay', 'EndDay','Duration', 'MinSpell','EventsPerYear','MaxInter-event','FlowLevelVolume','HighReleaseWindowStart', 'HighReleaseWindowEnd', 'LowReleaseWindowStart', 'LowReleaseWindowEnd','PeakLevelWindowStart', 'PeakLevelWindowEnd', 'LowLevelWindowStart', 'LowLevelWindowEnd','LevelThresholdMin','LevelThresholdMax']
    elif category == 'flow-ctf':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration', 'MinSpell', 'WithinEventGapTolerance', 'EventsPerYear', 'MaxInter-event', 'FlowLevelVolume', 'NonFlowSpell', 'CtfThreshold']
    elif category == 'rise_fall':
        pull = ['StartMonth', 'EndMonth', 'StartDay', 'EndDay', 'FlowThresholdMin', 'FlowThresholdMax', 'Duration', 'MinSpell', 'WithinEventGapTolerance', 'EventsPerYear', 'MaxInter-event', 'FlowLevelVolume', 'NonFlowSpell', 'RateOfRiseMax1', 'RateOfRiseMax2', 'RateOfFallMin', 'RateOfRiseThreshold1', 'RateOfRiseThreshold2', 'RateOfRiseRiverLevel', 'RateOfFallRiverLevel' ]
    return pull

def get_bad_QA_codes() -> list:
    '''NSW codes representing poor quality data.
    
    Returns:
        list: quality codes needing to be filtered out.

    '''
    return [151, 152, 153, 155, 180, 201, 202, 204, 205, 207, 223, 255]

def weirpool_type(ewr: str) -> str:
    '''Returns the type of Weirpool ewr. Currently only WP2 EWRs are classified as weirpool raisings
    
    Args:
        ewr (str): WP2 is considered raising, the remaining WP EWRs are considered falling

    Returns:
        str: either 'raising' or 'falling'
    
    '''

    return 'raising' if ewr == 'WP2' else 'falling'

@cached(cache=TTLCache(maxsize=1024, ttl=1800))
def get_planning_unit_info() -> pd.DataFrame:
    '''Run this function to get the planning unit MDBA ID and equivilent planning unit name as specified in the LTWP.
    
    Result:
        pd.DataFrame: dataframe with planning units and their unique planning unit ID.
    
    '''
    EWR_table = get_EWR_table()
        
    planningUnits = EWR_table.groupby(['PlanningUnitID', 'PlanningUnitName']).size().reset_index().drop([0], axis=1) 
    
    return planningUnits



# Function to pull out the ewr parameter information
def ewr_parameter_grabber(EWR_TABLE: pd.DataFrame, gauge: str, pu: str, ewr: str, PARAMETER: str) -> str:
    '''
    Input an ewr table to pull data from, a gauge, planning unit, and ewr for the unique value, and a requested parameter

    Args:
        EWR_TABLE (pd.DataFrame): dataset of EWRs
        gauge (str): Gauge string
        pu (str): Planning unit name
        ewr (str): ewr string
        PARAMETER (str): which parameter of the ewr to access
    Results:
        str: requested ewr component
    
    '''
    component = (EWR_TABLE[((EWR_TABLE['Gauge'] == gauge) & 
                           (EWR_TABLE['Code'] == ewr) &
                           (EWR_TABLE['PlanningUnitName'] == pu)
                          )][PARAMETER]).to_list()[0]
    return component if component else 0

def get_barrage_flow_gauges()-> dict:
    """Returns a dictionary of the flow gauges associated with each barrage.
    Results:
        dict: dictionary of flow gauges associated with each barrage.       	
    """

    flow_barrage_gauges = {'A4261002': ['A4261002']}

    return flow_barrage_gauges

def get_barrage_level_gauges()-> dict:
    """Returns a dictionary of the level gauges associated with each barrage.
    Results:
        dict: dictionary of level gauges associated with each barrage.
    """

    level_barrage_gauges = {'A4260524': ['A4260527','A4261133', 'A4260524', 'A4260574', 'A4260575' ],
                            'A4260633' : ['A4260633','A4261209', 'A4261165']}
    
    return level_barrage_gauges

def get_qld_level_gauges()-> list:
    """Returns a dictionary of the level gauges associated with each barrage.
    Results:
        dict: dictionary of level gauges associated with each barrage.
    """
    return ['422015','422030', '422034', '416011', '416048']

def get_qld_flow_gauges()-> list:
    """Returns a dictionary of the flow gauges associated with each barrage.
    Results:
        dict: dictionary of flow gauges associated with each barrage.
    """
    return ['422030','422207A',
            '422209A',
            '422211A',
            '422502A',
            '424201A',
            '422034' ]

def get_vic_level_gauges()-> list:
    """Returns a list of the level gauges for VIC.
    Results:
        dict: dictionary of flow gauges associated with each barrage.
    """
    return ['405201', '405202', '405200']


def get_cllmm_gauges()->list:
    return ["A4261002", "A4260524", "A4260633"]


def get_gauges(category: str, ewr_table_path: str = None) -> set:
    '''
    Gathers a list of either all gauges that have EWRs associated with them,
    a list of all flow type gauges that have EWRs associated with them,
    or a list of all level type gauges that have EWRs associated with them

    Args:
        category(str): options = 'all gauges', 'flow gauges', or 'level gauges'
    Returns:
        list: list of gauges in selected category.

    '''
    EWR_table = get_EWR_table(file_path=ewr_table_path)
    menindee_gauges, wp_gauges = get_level_gauges()
    wp_gauges = list(wp_gauges.values())
    flow_barrage_gauges = [ val for sublist in get_barrage_flow_gauges().values() for val in sublist]
    level_barrage_gauges = [ val for sublist in get_barrage_level_gauges().values() for val in sublist]
    qld_level_gauges = get_qld_level_gauges()
    qld_flow_gauges = get_qld_flow_gauges()
    vic_level_gauges = get_vic_level_gauges()
    
    multi_gauges = get_multi_gauges('gauges')
    multi_gauges = list(multi_gauges.values())
    if category == 'all gauges':
        return set(EWR_table['Gauge'].to_list()+menindee_gauges+wp_gauges+multi_gauges+flow_barrage_gauges+level_barrage_gauges+qld_flow_gauges+qld_level_gauges+vic_level_gauges)
    elif category == 'flow gauges':
        return set(EWR_table['Gauge'].to_list() + multi_gauges + flow_barrage_gauges + qld_flow_gauges) 
    elif category == 'level gauges':
        level_gauges = EWR_table[EWR_table['FlowLevelVolume']=='L']['Gauge'].to_list()
        return set(menindee_gauges + wp_gauges + level_barrage_gauges + qld_level_gauges + level_gauges + vic_level_gauges)
    else:
        raise ValueError('''No gauge category sent to the "get_gauges" function''')
    
def get_scenario_gauges(gauge_results: dict) -> list:
    """return a list of gauges process for the scenatios(s)

    Args:
        gauge_results (dict): the dictionary of gauge results either for
        observed or scenarios handlers

    Returns:
        list: list of gauges process for the scenatios(s)
    """
    scenario_gauges = []
    for scenario in gauge_results.values():
        for gauge in scenario.keys():
            scenario_gauges.append(gauge)
    return list(set(scenario_gauges))


def gauge_groups(parameter_sheet: pd.DataFrame) -> dict:
    '''
    Returns a dictionary of flow, level, and lake level gauges based on the parameter sheet and some hard coding of other EWRs

    Args:
        parameter_sheet (pd.DataFrame): input parameter sheet
    
    Returns:
        dict: keys as flow, level, and lake level gauges, values as the list of gauges
    '''
    
    # Hard coded gauges for the CLLMM EWRs
    hard_code_levels = ['A4260527', 'A4260524', 'A4260633', 'A4261209', 'A4261165']
    hard_code_lake_levels = ['A4261133', 'A4260574', 'A4260575']

    flow_gauges = set(parameter_sheet[parameter_sheet['GaugeType'] == 'F']['Gauge']) + set(parameter_sheet['Multigauge'])
    level_gauges = set(parameter_sheet[parameter_sheet['GaugeType'] == 'L']['Gauge']) + set(parameter_sheet['WeirpoolGauge']) + set(hard_code_levels)
    lake_level_gauges = set(parameter_sheet[parameter_sheet['GaugeType'] == 'LL']['Gauge'])+set(hard_code_lake_levels)

    return flow_gauges, level_gauges, lake_level_gauges

# def gauges_to_measurand()


def get_obj_mapping(
    parameter_sheet_path=None,
    objective_reference_path=None
) -> pd.DataFrame:
    '''
    Retrieves objective mapping and merges with parameter sheet on Gauge, Planning Unit, Code, and LTWPShortName
    to create a longform table that lists the associations between each EWR and ecological objectives.
    Returns a DataFrame with the merged result.
    Args:
        param_sheet_path (str) = file path to parameter sheet. If None, default parameter_sheet.csv inside the tool is selected
        obj_ref_path (str) = file path to objective mapping csv. If Not, default objective_reference.csv inside EWR tool is selected
    '''
    param_sheet_cols = [
        'PlanningUnitName', 'LTWPShortName',  'SWSDLName', 'State', 'Gauge', 'Code', 'EcoObj'
    ]

    if not objective_reference_path:
        objective_reference_path = os.path.join(BASE_PATH, "parameter_metadata/obj_reference.csv")
    

    obj_ref = pd.read_csv(objective_reference_path)
    

    df = get_EWR_table(file_path=parameter_sheet_path)
    df_sub = df[param_sheet_cols]
    
    # Split 'EnvObj' by '+' and explode to long format
    longform_ewr = df_sub.assign(
        EnvObj=df_sub['EcoObj'].str.split('+')
    ).explode('EnvObj').drop_duplicates()

    merged_df = longform_ewr.merge(
        obj_ref,
        left_on= ['LTWPShortName', 'PlanningUnitName', 'Gauge', 'Code', 'EcoObj', 'SWSDLName', 'State'],
        right_on=['LTWPShortName', 'PlanningUnitName', 'Gauge', 'Code', 'EcoObj', 'SWSDLName', 'State'],
        how='left'
    )

    return merged_df



from pathlib import Path
import io

import pandas as pd
from pandas._testing import assert_frame_equal
import numpy as np
import requests
from datetime import datetime
import os
import random
import string
import pytest
import re
import json

from py_ewr import data_inputs


BASE_PATH = Path(__file__).resolve().parents[1]   

parameter_sheet_path = os.path.join(BASE_PATH, "py_ewr/parameter_metadata/parameter_sheet.csv")
    
def test_get_multi_gauges():
    '''
    1. Test for returning planning units and gauges where there are multi gauge ewr requirements
    2. Test for returning the unique gauge to gauge dictionaries where there are multi gauge ewr requirements
    '''
    # Test 1
    expected_multi_gauges = {'PU_0000130': {'421090': '421088', '421088': '421090'},
              'PU_0000131': {'421090': '421088', '421088': '421090'},
              'PU_0000132': {'421090': '421088', '421088': '421090'},
              'PU_0000133': {'421090': '421088', '421088': '421090'},
              'PU_0000251': {'423001': '423002'},
              'PU_0000280': {'1AS': '1ES'}
             }
    multi_gauges = data_inputs.get_multi_gauges('all')
    assert expected_multi_gauges == multi_gauges
    #--------------------------------------------------------
    # Test 2
    expected_multi_gauges = {'421090': '421088', '421088': '421090', '423001': '423002', '1AS': '1ES'}
    multi_gauges = data_inputs.get_multi_gauges('gauges')
    assert expected_multi_gauges == multi_gauges    

def test_get_EWR_table():
    '''
    - 1. Test for equal entries (no lost EWRs)
    - 2. Test to ensure no bad EWRs make it through using a subset of EWRs
    - 3. Test EWR table columns match the expected columns
    -4. Test error thrown when startMonth and endMonth are not found in the table columns
    '''
    # Test 1
    proxies={} # Populate with your proxy settings

    cols=('PlanningUnitID', 'PlanningUnitName',  'LTWPShortName', 'CompliancePoint/Node', 'Gauge', 'Code',
                                'TargetFrequency', 'TargetFrequencyMin', 'TargetFrequencyMax', 'EventsPerYear', 'Duration', 'MinSpell', 
                                'FlowThresholdMin', 'FlowThresholdMax', 'MaxInter-event', 'WithinEventGapTolerance', 'WeirpoolGauge', 'FlowLevelVolume', 
                                'LevelThresholdMin', 'LevelThresholdMax', 'VolumeThreshold', 'DrawdownRate', 'AccumulationPeriod',
                                'Multigauge', 'MaxSpell', 'TriggerDay', 'TriggerMonth', 'DrawDownRateWeek', 'CtfThreshold','NonFlowSpell', 'MaxLevelChange')
    essential_cols = ('StartMonth', 'EndMonth')

    comb_cols = cols + essential_cols

    df = pd.read_csv(parameter_sheet_path,
                    usecols = comb_cols,
                     dtype='str', encoding='cp1252'
                    )
    
    # Get the cleaned dataset:
    # testing 1 and 2 
    EWR_table = data_inputs.get_EWR_table()

    assert len(df), len(EWR_table)

    #test 3 
    EWR_table = data_inputs.get_EWR_table(file_path = parameter_sheet_path, columns_to_keep = comb_cols)
    assert sorted(EWR_table.columns.tolist()) == sorted(comb_cols+('StartDay', 'EndDay'))

    # test 4 if start month and end month are not in the parmaeter sheet raise error
    
    with pytest.raises(KeyError):
        EWR_table = data_inputs.get_EWR_table(file_path = parameter_sheet_path, columns_to_keep = cols)

def test_get_ewr_calc_config():
    '''
    1. Test for correct return of ewr calculation config
    assert it returns a dictionary
    '''

    ewr_calc_config = data_inputs.get_ewr_calc_config()

    assert isinstance(ewr_calc_config, dict)
    assert "flow_handle" in ewr_calc_config.keys()


def test_get_barrage_flow_gauges():
    result = data_inputs.get_barrage_flow_gauges()
    assert isinstance(result, dict)
    # there is only one key in the dictionary
    assert len(result) == 1
    ## all key main gauge belong to the list of gauges
    for k, v in result.items():
        assert isinstance(v, list)
        assert k in v


def test_get_barrage_level_gauges():
    result = data_inputs.get_barrage_level_gauges()
    assert isinstance(result, dict)
    # there is only one key in the dictionary
    ## all key main gauge belong to the list of gauges
    for k, v in result.items():
        assert isinstance(v, list)
        assert k in v


def test_get_cllmm_gauges():
    result = data_inputs.get_cllmm_gauges()
    assert isinstance(result, list)
    for item in result:
        assert isinstance(item, str)

@pytest.mark.parametrize('expected_results', 
[
    (
    ['A4260527', 'A4260633', 'A4260634', 'A4260635', 'A4260637', 'A4261002']
    )
])
def test_get_scenario_gauges(gauge_results, expected_results):
    result = data_inputs.get_scenario_gauges(gauge_results)
    assert sorted(result) == expected_results
    
def test_get_iqqm_codes():
    result = data_inputs.get_iqqm_codes()
    stations = {
        '229': '421023',
        '42': '421001',
        '464': '421011',
        '240': '421019',
        '266': '421146',
        '951': '421090',
        '487': '421022',
        '130': '421012',
        '171': '421004',
    }
    assert stations == result


def generate_years(start, end, leap = False):
    if leap == True:
        leaps =  [year for year in range(start, end + 1) \
                if (year % 4 == 0 and year % 100 != 0) or (year % 400 == 0)]
        rand_year = random.choice(leaps)
    else:
        non_leaps = [year for year in range(start, end + 1) \
                if (year % 4 != 0 and year % 100 == 0) or (year % 400 != 0)]
        rand_year = random.choice(non_leaps)
    return rand_year

def check_EWR_logic(df: pd.DataFrame, year: int):
    '''
    Checks the logic between columns that specify aspects of the flow regime related to timing
    1. DURATION CHECK: 
        Checks that duration <= days between start and end month
    2. EVENT NUMBER CHECK: 
        Checks that the sum of events*event number per year <= 365 days.
    3. minSpell is not greater than duration
    4. FLOW THRESHOLD CHECK: 
        Checks: minimum flow threshold  =< maximum flow threshold
    5. TARGET FREQUENCY CHECK:
        Checks minimum target frequency =< target_frequency >= maximum target frequency
    6. LEVEL THRESHOLD CHECK
        Check minimum level threshold =< maximum level threshold. 
    7. DUPLICATE ROW CHECK
        Check unique combinations of planning unit, EWR,and gauge occur only once in the dataset
    8. SPECIAL CHARACTER CHECK
        Checks if the dataframe is free of special characters.
    9. NO THRESHOLD CHECK
        Check that all EWRs contain at least one flow or volume threshold value either minimum or maximum
    10. NO DURATION CHECK
        Check that all EWRs contain a duration value
    11. DSF EWRS
        Check that EWRs str contains DSF are not present in the final dataframe

    args: df: an EWR table as a dataframe
         year: an integer year
    
    return: 
        number of rows per dataset that violate the conditions described above.
    '''
    checking_df = df.copy()

    # Convert numeric columns to float
    columns = ['TargetFrequency', 'TargetFrequencyMin', 'TargetFrequencyMax',
           'EventsPerYear', 'Duration', 'MinSpell', 'FlowThresholdMin',
           'FlowThresholdMax', 'LevelThresholdMin', 'LevelThresholdMax']
    checking_df[columns] = checking_df[columns].replace({' ': np.nan, '': np.nan}, regex=True)
    checking_df[columns] = checking_df[columns].astype(float)
    
    # Duration check
    dur_filter = checking_df.copy()
    
    date_cols = ['StartMonth', 'StartDay', 'EndMonth', 'EndDay']

    # Convert all date columns: handle NaN and zeros
    for col in date_cols:
        dur_filter[col] = pd.to_numeric(dur_filter[col], errors='coerce').fillna(1).replace(0, 1).astype(int)
    
    # # Filter out July-June water years
    # dur_filter = dur_filter[~((dur_filter['StartMonth'] == 7) & (dur_filter["EndMonth"] == 6))]

    # Create date columns
    dur_filter['StartDate'] = pd.to_datetime(dur_filter.apply(
        lambda row: f"{year}-{row['StartMonth']:02d}-{row['StartDay']:02d}", axis=1))

    dur_filter['EndDate'] = pd.to_datetime(dur_filter.apply(
        lambda row: f"{year+1 if row['StartMonth'] > row['EndMonth'] else year}-{row['EndMonth']:02d}-{row['EndDay']:02d}", axis=1)) + pd.offsets.MonthEnd(1)

    dur_filter['DaysBetween'] = (dur_filter['EndDate'] - dur_filter['StartDate']).dt.days + 1


    # remove cease to flows since they don't follow the same rules
    dur_filter_cf_removed = dur_filter[~dur_filter['Code'].str.contains('CF|CTF')]
    
    # Duration violation check with NaN handling
    duration_violation = dur_filter_cf_removed[
        (dur_filter_cf_removed['Duration'] > dur_filter_cf_removed['DaysBetween']) &
        dur_filter_cf_removed['Duration'].notna() &
        dur_filter_cf_removed['DaysBetween'].notna()
    ]

    checking_df = checking_df[~checking_df['Code'].str.contains('CF|CTF')]

    # Calculate MaxEventDays
    checking_df['MaxEventDays'] = checking_df['EventsPerYear'] * checking_df['Duration']

    # Event Number Check with NaN handling
    event_number_violation = checking_df[
        (checking_df['MaxEventDays'] > 365) &
        checking_df['MaxEventDays'].notna()
    ]

    # MinSpell Check with NaN handling
    min_spell_violation = checking_df[
        (checking_df['MinSpell'] > checking_df['Duration']) &
        checking_df['MinSpell'].notna() &
        checking_df['Duration'].notna()
    ]

    # Flow Threshold Check with NaN handling
    flow_threshold_violation = checking_df[
        (checking_df['FlowThresholdMax'] > 0) & 
        (checking_df['FlowThresholdMin'] > checking_df['FlowThresholdMax']) &
        checking_df['FlowThresholdMin'].notna() &
        checking_df['FlowThresholdMax'].notna()
    ]

    # Level Threshold Check with NaN handling
    level_threshold_violation = checking_df[
        (checking_df['LevelThresholdMax'] > 0) & 
        (checking_df['LevelThresholdMin'] > checking_df['LevelThresholdMax']) &
        checking_df['LevelThresholdMin'].notna() &
        checking_df['LevelThresholdMax'].notna()
    ]

    # Target Frequency Check with NaN handling
    target_frequency_violation = checking_df[
        (checking_df['TargetFrequencyMax'] > 0) & 
        (
            (checking_df['TargetFrequencyMin'] >= checking_df['TargetFrequency']) &
            (checking_df['TargetFrequency'] >= checking_df['TargetFrequencyMax']) &
            (checking_df['TargetFrequencyMin'] >= checking_df['TargetFrequencyMax'])
        ) &
        checking_df['TargetFrequency'].notna() &
        checking_df['TargetFrequencyMin'].notna() &
        checking_df['TargetFrequencyMax'].notna()
    ]

    # Duplicate EWR planning units and gauges 
    checking_df['unique_ID'] = checking_df['Gauge'] + '_' + checking_df['PlanningUnitID'] + '_' + checking_df['Code']
    duplicates = checking_df[checking_df.duplicated('unique_ID', keep=False)]


    # additional conditions
    #  EWRs with 'See notes' in start and month
    see_notes = checking_df[(checking_df["StartMonth"] == 'See note') & (checking_df["EndMonth"] == 'See note')]

    # # Filtering those with no flow/level/volume thresholds
    no_thresh = checking_df[(checking_df["FlowThresholdMin"] == '') & \
                    (checking_df["FlowThresholdMax"] == '') &\
                    (checking_df["VolumeThreshold"] == '') &\
                    (checking_df["LevelThresholdMin"] == '') &\
                    (checking_df["LevelThresholdMax"] == '')]

    # # Filtering those with no durations
    no_duration = checking_df[checking_df["Duration"] == '']

    # # Filtering DSF EWRs
    DSF_EWRs = checking_df[checking_df['Code'].str.startswith('DSF')]
   

     # Special Character Check
    allowed_chars = string.ascii_letters + string.digits+',()+-_./:%@ '
    pattern = f'^[{re.escape(allowed_chars)}]*$'
    
    # Apply the pattern to filter the DataFrame
    spec_char = df[~df.apply(lambda x: x.astype(str).str.match(pattern) | x.isna()).all(axis=1)]

    # print the issues

    if not duration_violation.empty:
        print('#------------- Duration Violation -------------#')
        print(duration_violation[['Gauge', 'Code', 'LTWPShortName', 'Duration', 'DaysBetween']])
    if not no_duration.empty:
        print('#------------- Empty duration -------------#')
        print(no_duration[['Gauge', 'Code', 'LTWPShortName', 'Duration']])
    if not event_number_violation.empty:
        print('#------------- Event number over length of seasonal window -------------#')
        print(event_number_violation[['Gauge', 'Code', 'LTWPShortName', 'MaxEventDays', 'EventsPerYear', 'Duration']])
    if not min_spell_violation.empty:
        print('#------------- Minimum Spell > Duration -------------#')
        print(min_spell_violation[['Gauge', 'Code', 'LTWPShortName', 'MinSpell', 'Duration']])
    if not flow_threshold_violation.empty:
        print('#------------- Flow Threshold min max relationships not correct -------------#')
        print(flow_threshold_violation[['Gauge', 'Code', 'LTWPShortName', 'FlowThresholdMin', 'FlowThresholdMax']])
    if not level_threshold_violation.empty:
        print('#------------- Level Threshold min max relationships not correct -------------#')
        print(level_threshold_violation[['Gauge', 'Code', 'LTWPShortName', 'LevelThresholdMin', 'LevelThresholdMax']])
    if not target_frequency_violation.empty:
        print('#-------------  Target Frequency relationships not correct -------------#')
        print(target_frequency_violation[['Gauge', 'Code', 'LTWPShortName', 'TargetFrequency', 'TargetFrequencyMin', 'TargetFrequencyMax']])
    if not duplicates.empty:
        print('#------------- Duplicate rows -------------#')
        print(duplicates[['Gauge', 'PlanningUnitID', 'LTWPShortName', 'Code']])
    if not see_notes.empty:
        print("#------------- EWRs with 'see notes' in start or end month columns and therefore will not be calculated -------------#")
        print(see_notes[['Gauge', 'PlanningUnitID', 'LTWPShortName', 'Code']])
    if not no_thresh.empty:
        print('#------------- EWRs with no flow or volume thresholds -------------#')
        print(no_thresh[['Gauge', 
                         'PlanningUnitID', 
                         'LTWPShortName', 
                         'Code', 
                         'FlowThresholdMin', 
                         'FlowThresholdMax',
                         'VolumeThreshold', 
                         'LevelThresholdMin',
                         'LevelThresholdMax"']])
    if not DSF_EWRs.empty:
        print(DSF_EWRs[['Gauge', 'PlanningUnitID', 'LTWPShortName', 'Code']])
    if not spec_char.empty:
        print('#------------- Special characters in the following rows -------------#')
        print(spec_char)

     # Check if there are no violations
    no_violations = all(len(v) == 0 for v in [
        duration_violation,
        no_duration,
        event_number_violation,
        min_spell_violation,
        flow_threshold_violation,
        level_threshold_violation,
        target_frequency_violation,
        duplicates,
        see_notes, 
        no_thresh, 
        DSF_EWRs,
        spec_char
    ])
    assert no_violations, "Errors were found with the logic in the EWR table"
    

def test_check_EWR_logic():
    non_leap = generate_years(1910, 2024)
    leap = generate_years(1910, 2024, True)
    parameter_sheet_path = os.path.join(BASE_PATH, "py_ewr/parameter_metadata/parameter_sheet.csv")
    columns_to_keep = ('PlanningUnitID', 'LTWPShortName', 'PlanningUnitName', 'Gauge', 'Code', 'StartMonth', 'TargetFrequency', 'TargetFrequencyMin', 'TargetFrequencyMax', 'State', 'SWSDLName',
                          'EndMonth', 'EventsPerYear', 'Duration', 'MinSpell', 
                          'FlowThresholdMin', 'FlowThresholdMax', 'MaxInter-event', 'WithinEventGapTolerance', 'WeirpoolGauge', 'FlowLevelVolume', 
                          'LevelThresholdMin', 'LevelThresholdMax', 'VolumeThreshold', 'DrawdownRate', 'MaxLevelChange', 'AccumulationPeriod',
                          'Multigauge', 'MaxSpell', 'TriggerDay', 'TriggerMonth', 'DrawDownRateWeek','AnnualBarrageFlow',
                          'ThreeYearsBarrageFlow', 'HighReleaseWindowStart', 'HighReleaseWindowEnd', 'LowReleaseWindowStart', 'LowReleaseWindowEnd',
                          'PeakLevelWindowStart', 'PeakLevelWindowEnd', 'LowLevelWindowStart', 'LowLevelWindowEnd', 'NonFlowSpell','EggsDaysSpell',
                          'LarvaeDaysSpell', 'RateOfRiseMax1','RateOfRiseMax2','RateOfFallMin','RateOfRiseThreshold1',
                          'RateOfRiseThreshold2','RateOfRiseRiverLevel','RateOfFallRiverLevel', 'CtfThreshold', 'GaugeType')
    
    EWR_table = data_inputs.get_EWR_table(parameter_sheet_path, columns_to_keep)
    check_EWR_logic(EWR_table, non_leap)
    check_EWR_logic(EWR_table, leap)

@pytest.mark.parametrize('test_id,  test_data, expected_start_month, expected_start_day, expected_end_month,expected_end_day', 
    [ 
        (
           'test case 1', 
            {
                'Gauge': ['409025'],
                'Code': ['nestS1'],
                'StartMonth': ['9.15'],  # First has day, second doesn't
                'EndMonth': ['11.15']
            },
            9, 15, 11, 15
        )
        ,

        (
           'test case 2', 
            {
                'Gauge': ['409025'],
                'Code': ['BF3'],
                'StartMonth': ['7'],  # second doesn't have a specific day
                'EndMonth': ['6']
            },
            7, None, 6, None
        )
    ] 
)

def test_modify_EWR_table(
    test_id, 
    test_data,
    expected_start_month,
    expected_start_day,
    expected_end_month,
    expected_end_day 
):
    '''
    Test that datatypes are coerced to the correct datatype
    1. Test a float is separated into StartDay and StartMonth columns (e.g., "9.15" -> month=9, day=15)
    2. Test integer remains integer (e.g., "7" -> month=7, day=None)
    '''
    # Get the EWR table
    
    test_df = pd.DataFrame(test_data)
    
    # Apply the modify function
    result = data_inputs.modify_EWR_table(test_df)

    row = result.iloc[0]
    
    # Test StartMonth
    assert row['StartMonth'] == expected_start_month, \
        f"StartMonth mismatch for {test_id}: expected {expected_start_month}, got {row['StartMonth']}"
    
    # Test StartDay
    if expected_start_day is None:
        assert pd.isna(row['StartDay']) or row['StartDay'] is None, \
            f"StartDay should be None/NaN for {test_id}, got {row['StartDay']}"
    else:
        assert row['StartDay'] == expected_start_day, \
            f"StartDay mismatch for {test_id}: expected {expected_start_day}, got {row['StartDay']}"
    
    # Test EndMonth
    assert row['EndMonth'] == expected_end_month, \
        f"EndMonth mismatch for {test_id}: expected {expected_end_month}, got {row['EndMonth']}"
    
    # Test EndDay
    if expected_end_day is None:
        assert pd.isna(row['EndDay']) or row['EndDay'] is None, \
            f"EndDay should be None/NaN for {test_id}, got {row['EndDay']}"
    else:
        assert row['EndDay'] == expected_end_day, \
            f"EndDay mismatch for {test_id}: expected {expected_end_day}, got {row['EndDay']}"

def test_modify_EWR_table_datatypes():
    '''
    1. Test that the dtypes of the dataframe are as expected after running through function
    '''
    # Get the EWR table
    test_df = pd.read_csv(parameter_sheet_path, dtype = 'str')
    
    # Apply the modify function
    result = data_inputs.modify_EWR_table(test_df)

    expected_dtypes = {
        # Integer columns
        'AccumulationPeriod': 'Int64',
        'AnnualBarrageFlow': 'Int64',
        'Duration': 'Int64',
        'EggsDaysSpell': 'Int64',
        'EndDay': 'Int64',
        'EndMonth': 'Int64',
        'EventsPerYear': 'Int64',
        'FlowThresholdMax': 'Int64',
        'FlowThresholdMin': 'Int64',
        'HighReleaseWindowEnd': 'Int64',
        'HighReleaseWindowStart': 'Int64',
        'LarvaeDaysSpell': 'Int64',
        'LowLevelWindowEnd': 'Int64',
        'LowLevelWindowStart': 'Int64',
        'LowReleaseWindowEnd': 'Int64',
        'LowReleaseWindowStart': 'Int64',
        'MaxSpell': 'Int64',
        'MinSpell': 'Int64',
        'NonFlowSpell': 'Int64',
        'PeakLevelWindowEnd': 'Int64',
        'PeakLevelWindowStart': 'Int64',
        'StartDay': 'Int64',
        'StartMonth': 'Int64',
        'ThreeYearsBarrageFlow': 'Int64',
        'TriggerDay': 'Int64',
        'TriggerMonth': 'Int64',
        'VolumeThreshold': 'Int64',
        'WithinEventGapTolerance': 'Int64',
        
        # Float columns
        'CtfThreshold': 'Float64',
        'DrawDownRateWeek': 'Float64',
        'LevelThresholdMax': 'Float64',
        'LevelThresholdMin': 'Float64',
        'MaxInter-event': 'Float64',
        'MaxLevelChange': 'Float64',
        'RateOfFallMin': 'Float64',
        'RateOfFallRiverLevel': 'Float64',
        'RateOfRiseMax1': 'Float64',
        'RateOfRiseMax2': 'Float64',
        'RateOfRiseRiverLevel': 'Float64',
        'RateOfRiseThreshold1': 'Float64',
        'RateOfRiseThreshold2': 'Float64'
    }

    keys = expected_dtypes.keys()

    result_key_cols = result[keys]

    assert result_key_cols.dtypes.astype(str).to_dict() == expected_dtypes, f"expected datatypes not matching output datatype"
    




    
